@@ -26,23 +26,17 @@ import java.util.List;
 public class ConsoleReader implements ConsoleOperations {
 
     final static int TAB_WIDTH = 4;
-
     String prompt;
-
     private boolean useHistory = true;
-
     private boolean usePagination = false;
-
     public static final String CR = System.getProperty("line.separator");
-
-    private static ResourceBundle loc = ResourceBundle
-            .getBundle(CandidateListCompletionHandler.class.getName());
-
+    private static ResourceBundle loc = ResourceBundle.getBundle(CandidateListCompletionHandler.class.getName());
     /**
      * Map that contains the operation name to keymay operation mapping.
      */
     public static SortedMap KEYMAP_NAMES;
 
+
     static {
         Map names = new TreeMap();
 
@@ -93,30 +87,26 @@ public class ConsoleReader implements ConsoleOperations {
         names.put("COMPLETE", new Short(COMPLETE));
         names.put("EXIT", new Short(EXIT));
         names.put("CLEAR_LINE", new Short(CLEAR_LINE));
+        names.put("ABORT", new Short(ABORT));
 
         KEYMAP_NAMES = new TreeMap(Collections.unmodifiableMap(names));
     }
-
     /**
      * The map for logical operations.
      */
     private final short[] keybindings;
-
     /**
      * If true, issue an audible keyboard bell when appropriate.
      */
     private boolean bellEnabled = true;
-
     /**
      * The current character mask.
      */
     private Character mask = null;
-
     /**
      * The null mask.
      */
     private static final Character NULL_MASK = new Character((char) 0);
-
     /**
      * The number of tab-completion candidates above which a warning will be
      * prompted before showing all the candidates.
@@ -125,45 +115,39 @@ public class ConsoleReader implements ConsoleOperations {
             "jline.completion.threshold", 100).intValue(); // same default as
 
     // bash
-
     /**
      * The Terminal to use.
      */
     private final Terminal terminal;
-
     private CompletionHandler completionHandler = new CandidateListCompletionHandler();
-
     InputStream in;
-
     final Writer out;
-
     final CursorBuffer buf = new CursorBuffer();
-
     static PrintWriter debugger;
-
     History history = new History();
-
     final List completors = new LinkedList();
-
     private Character echoCharacter = null;
+    private Map triggeredActions = new HashMap();
 
-	private Map triggeredActions = new HashMap();
+    private StringBuffer searchTerm = null;
+    private String previousSearchTerm = "";
+    private int searchIndex = -1;
+
+    /**
+     * Adding a triggered Action allows to give another course of action
+     * if a character passed the preprocessing.
+     *
+     * Say you want to close the application if the user enter q.
+     * addTriggerAction('q', new ActionListener(){ System.exit(0); });
+     * would do the trick.
+     *
+     * @param c
+     * @param listener
+     */
+    public void addTriggeredAction(char c, ActionListener listener) {
+        triggeredActions.put(new Character(c), listener);
+    }
 
-	/**
-	 * Adding a triggered Action allows to give another curse of action
-	 * if a character passed the preprocessing.
-	 * 
-	 * Say you want to close the application if the user enter q.
-	 * addTriggerAction('q', new ActionListener(){ System.exit(0); });
-	 * would do the trick.
-	 * 
-	 * @param c
-	 * @param listener
-	 */
-	public void addTriggeredAction(char c, ActionListener listener){
-		triggeredActions.put(new Character(c), listener);
-	}
-	
     /**
      * Create a new reader using {@link FileDescriptor#in} for input and
      * {@link System#out} for output. {@link FileDescriptor#in} is used because
@@ -171,7 +155,9 @@ public class ConsoleReader implements ConsoleOperations {
      */
     public ConsoleReader() throws IOException {
         this(new FileInputStream(FileDescriptor.in),
-                new PrintWriter(System.out));
+                new PrintWriter(
+                new OutputStreamWriter(System.out,
+                System.getProperty("jline.WindowsTerminal.output.encoding", System.getProperty("file.encoding")))));
     }
 
     /**
@@ -205,16 +191,15 @@ public class ConsoleReader implements ConsoleOperations {
         this.terminal = term;
         setInput(in);
         this.out = out;
-
         if (bindings == null) {
             try {
                 String bindingFile = System.getProperty("jline.keybindings",
-                    new File(System.getProperty("user.home",
-                        ".jlinebindings.properties")).getAbsolutePath());
+                    new File(System.getProperty("user.home"),
+                        ".jlinebindings.properties").getAbsolutePath());
 
                 if (new File(bindingFile).isFile()) {
                     bindings = new FileInputStream(new File(bindingFile));
-                } 
+                }
             } catch (Exception e) {
                 // swallow exceptions with option debugging
                 if (debugger != null) {
@@ -258,11 +243,11 @@ public class ConsoleReader implements ConsoleOperations {
                 }
             }
 
-            // hardwired arrow key bindings
-            // keybindings[VK_UP] = PREV_HISTORY;
-            // keybindings[VK_DOWN] = NEXT_HISTORY;
-            // keybindings[VK_LEFT] = PREV_CHAR;
-            // keybindings[VK_RIGHT] = NEXT_CHAR;
+        // hardwired arrow key bindings
+        // keybindings[VK_UP] = PREV_HISTORY;
+        // keybindings[VK_DOWN] = NEXT_HISTORY;
+        // keybindings[VK_LEFT] = PREV_CHAR;
+        // keybindings[VK_RIGHT] = NEXT_CHAR;
         }
     }
 
@@ -329,7 +314,7 @@ public class ConsoleReader implements ConsoleOperations {
      * @return the width of the current terminal.
      */
     public int getTermwidth() {
-        return Terminal.setupTerminal().getTerminalWidth();
+        return getTerminal().getTerminalWidth();
     }
 
     /**
@@ -340,7 +325,7 @@ public class ConsoleReader implements ConsoleOperations {
      * @return the height of the current terminal.
      */
     public int getTermheight() {
-        return Terminal.setupTerminal().getTerminalHeight();
+        return getTerminal().getTerminalHeight();
     }
 
     /**
@@ -439,7 +424,33 @@ public class ConsoleReader implements ConsoleOperations {
     int getCursorPosition() {
         // FIXME: does not handle anything but a line with a prompt
         // absolute position
-        return ((prompt == null) ? 0 : prompt.length()) + buf.cursor;
+        return getStrippedAnsiLength(prompt) + buf.cursor;
+    }
+
+    /**
+     * Strips ANSI escape sequences starting with CSI and ending with char in range 64-126
+     * @param ansiString String possibly containing ANSI codes, may be null
+     * @return length after stripping ANSI codes
+     */
+    int getStrippedAnsiLength(String ansiString) {
+        if (ansiString ==  null) return 0;
+        boolean inAnsi = false;
+        int strippedLength = 0;
+        char[] chars = ansiString.toCharArray();
+        for (int i = 0; i < chars.length; i++) {
+            char c = chars[i];
+            if (!inAnsi && c == 27 && i < chars.length - 1 && chars[i+1] == '[') {
+                i++; // skip '['
+                inAnsi = true;
+            } else if (inAnsi) {
+                if (64 <= c && c <= 126) {
+                    inAnsi = false;
+                }
+            } else {
+                strippedLength++;
+            }
+        }
+        return strippedLength;
     }
 
     public String readLine(final String prompt) throws IOException {
@@ -472,8 +483,9 @@ public class ConsoleReader implements ConsoleOperations {
     public String readLine(final String prompt, final Character mask)
             throws IOException {
         this.mask = mask;
-        if (prompt != null)
+        if (prompt != null) {
             this.prompt = prompt;
+        }
 
         try {
             terminal.beforeReadLine(this, this.prompt, mask);
@@ -488,7 +500,14 @@ public class ConsoleReader implements ConsoleOperations {
                 return readLine(in);
             }
 
+            final int NORMAL = 1;
+            final int SEARCH = 2;
+            int state = NORMAL;
+
+            boolean success = true;
+
             while (true) {
+                // Read next key and look up the command binding.
                 int[] next = readBinding();
 
                 if (next == null) {
@@ -502,125 +521,219 @@ public class ConsoleReader implements ConsoleOperations {
                     return null;
                 }
 
-                boolean success = true;
+                // Search mode.
+                //
+                // Note that we have to do this first, because if there is a command
+                // not linked to a search command, we leave the search mode and fall
+                // through to the normal state.
+                if (state == SEARCH) {
+                    switch (code) {
+                        // This doesn't work right now, it seems CTRL-G is not passed
+                        // down correctly. :(
+                        case ABORT:
+                            state = NORMAL;
+                            break;
 
-                switch (code) {
-                case EXIT: // ctrl-d
+                        case SEARCH_PREV:
+                            if (searchTerm.length() == 0) {
+                                searchTerm.append(previousSearchTerm);
+                            }
 
-                    if (buf.buffer.length() == 0) {
-                        return null;
+                            if (searchIndex == -1) {
+                                searchIndex = history.searchBackwards(searchTerm.toString());
+                            } else {
+                                searchIndex = history.searchBackwards(searchTerm.toString(), searchIndex);
+                            }
+                            break;
+                            
+                        case DELETE_PREV_CHAR:
+                            if (searchTerm.length() > 0) {
+                                searchTerm.deleteCharAt(searchTerm.length() - 1);
+                                searchIndex = history.searchBackwards(searchTerm.toString());
+                            }
+                            break;
+                            
+                        case UNKNOWN:
+                            searchTerm.appendCodePoint(c);
+                            searchIndex = history.searchBackwards(searchTerm.toString());
+                            break;
+
+                        default:
+                            // Set buffer and cursor position to the found string.
+                            if (searchIndex != -1) {
+                                history.setCurrentIndex(searchIndex);
+                                setBuffer(history.current());
+                                buf.cursor = history.current().indexOf(searchTerm.toString());
+                            }
+                            state = NORMAL;
+                            break;
                     }
 
-                case COMPLETE: // tab
-                    success = complete();
-                    break;
-
-                case MOVE_TO_BEG:
-                    success = setCursorPosition(0);
-                    break;
-
-                case KILL_LINE: // CTRL-K
-                    success = killLine();
-                    break;
-
-                case CLEAR_SCREEN: // CTRL-L
-                    success = clearScreen();
-                    break;
-
-                case KILL_LINE_PREV: // CTRL-U
-                    success = resetLine();
-                    break;
-
-                case NEWLINE: // enter
-                    moveToEnd();
-                    printNewline(); // output newline
-                    return finishBuffer();
-
-                case DELETE_PREV_CHAR: // backspace
-                    success = backspace();
-                    break;
-
-                case DELETE_NEXT_CHAR: // delete
-                    success = deleteCurrentCharacter();
-                    break;
-
-                case MOVE_TO_END:
-                    success = moveToEnd();
-                    break;
-
-                case PREV_CHAR:
-                    success = moveCursor(-1) != 0;
-                    break;
-
-                case NEXT_CHAR:
-                    success = moveCursor(1) != 0;
-                    break;
-
-                case NEXT_HISTORY:
-                    success = moveHistory(true);
-                    break;
-
-                case PREV_HISTORY:
-                    success = moveHistory(false);
-                    break;
-
-                case REDISPLAY:
-                    break;
-
-                case PASTE:
-                    success = paste();
-                    break;
-
-                case DELETE_PREV_WORD:
-                    success = deletePreviousWord();
-                    break;
-
-                case PREV_WORD:
-                    success = previousWord();
-                    break;
-
-                case NEXT_WORD:
-                    success = nextWord();
-                    break;
-
-                case START_OF_HISTORY:
-                    success = history.moveToFirstEntry();
-                    if (success)
-                        setBuffer(history.current());
-                    break;
-
-                case END_OF_HISTORY:
-                    success = history.moveToLastEntry();
-                    if (success)
-                        setBuffer(history.current());
-                    break;
-
-                case CLEAR_LINE:
-                    moveInternal(-(buf.buffer.length()));
-                    killLine();
-                    break;
-
-                case INSERT:
-                    buf.setOvertyping(!buf.isOvertyping());
-                    break;
-
-                case UNKNOWN:
-                default:
-                    if (c != 0) { // ignore null chars
-                    	ActionListener action = (ActionListener) triggeredActions.get(new Character((char)c));
-                    	if (action != null)
-                    		action.actionPerformed(null);
-                    	else
-                    		putChar(c, true);
-                    } else
-                        success = false;
+                    // if we're still in search mode, print the search status
+                    if (state == SEARCH) {
+                        if (searchTerm.length() == 0) {
+                            printSearchStatus("", "");
+                        } else {
+                            if (searchIndex == -1) {
+                                beep();
+                            } else {
+                                printSearchStatus(searchTerm.toString(), history.getHistory(searchIndex));
+                            }
+                        }
+                    }
+                    // otherwise, restore the line
+                    else {
+                        restoreLine();
+                    }
                 }
 
-                if (!(success)) {
-                    beep();
-                }
+                if (state == NORMAL) {
+                    switch (code) {
+                        case EXIT: // ctrl-d
 
-                flushConsole();
+                            if (buf.buffer.length() == 0) {
+                                return null;
+                            }
+                            else {
+                                success = deleteCurrentCharacter();
+                            }
+                            break;
+
+                        case COMPLETE: // tab
+                            success = complete();
+                            break;
+
+                        case MOVE_TO_BEG:
+                            success = setCursorPosition(0);
+                            break;
+
+                        case KILL_LINE: // CTRL-K
+                            success = killLine();
+                            break;
+
+                        case CLEAR_SCREEN: // CTRL-L
+                            success = clearScreen();
+                            break;
+
+                        case KILL_LINE_PREV: // CTRL-U
+                            success = resetLine();
+                            break;
+
+                        case NEWLINE: // enter
+                            moveToEnd();
+                            printNewline(); // output newline
+                            return finishBuffer();
+
+                        case DELETE_PREV_CHAR: // backspace
+                            success = backspace();
+                            break;
+
+                        case DELETE_NEXT_CHAR: // delete
+                            success = deleteCurrentCharacter();
+                            break;
+
+                        case MOVE_TO_END:
+                            success = moveToEnd();
+                            break;
+
+                        case PREV_CHAR:
+                            success = moveCursor(-1) != 0;
+                            break;
+
+                        case NEXT_CHAR:
+                            success = moveCursor(1) != 0;
+                            break;
+
+                        case NEXT_HISTORY:
+                            success = moveHistory(true);
+                            break;
+
+                        case PREV_HISTORY:
+                            success = moveHistory(false);
+                            break;
+
+                        case ABORT:
+                        case REDISPLAY:
+                            break;
+
+                        case PASTE:
+                            success = paste();
+                            break;
+
+                        case DELETE_PREV_WORD:
+                            success = deletePreviousWord();
+                            break;
+
+                        case PREV_WORD:
+                            success = previousWord();
+                            break;
+
+                        case NEXT_WORD:
+                            success = nextWord();
+                            break;
+
+                        case START_OF_HISTORY:
+                            success = history.moveToFirstEntry();
+                            if (success) {
+                                setBuffer(history.current());
+                            }
+                            break;
+
+                        case END_OF_HISTORY:
+                            success = history.moveToLastEntry();
+                            if (success) {
+                                setBuffer(history.current());
+                            }
+                            break;
+
+                        case CLEAR_LINE:
+                            moveInternal(-(buf.buffer.length()));
+                            killLine();
+                            break;
+
+                        case INSERT:
+                            buf.setOvertyping(!buf.isOvertyping());
+                            break;
+
+                        case SEARCH_PREV: // CTRL-R
+                            if (searchTerm != null) {
+                                previousSearchTerm = searchTerm.toString();
+                            }
+                            searchTerm = new StringBuffer(buf.buffer);
+                            state = SEARCH;
+                            if (searchTerm.length() > 0) {
+                                searchIndex = history.searchBackwards(searchTerm.toString());
+                                if (searchIndex == -1) {
+                                    beep();
+                                }
+                                printSearchStatus(searchTerm.toString(),
+                                        searchIndex > -1 ? history.getHistory(searchIndex) : "");
+                            } else {
+                                searchIndex = -1;
+                                printSearchStatus("", "");
+                            }
+                            break;
+
+                        case UNKNOWN:
+                        default:
+                            if (c != 0) { // ignore null chars
+                                ActionListener action = (ActionListener) triggeredActions.get(new Character((char) c));
+                                if (action != null) {
+                                    action.actionPerformed(null);
+                                } else {
+                                    putChar(c, true);
+                                }
+                            } else {
+                                success = false;
+                            }
+                    }
+
+                    if (!(success)) {
+                        beep();
+                    }
+
+                    flushConsole();
+                }
             }
         } finally {
             terminal.afterReadLine(this, this.prompt, mask);
@@ -640,7 +753,7 @@ public class ConsoleReader implements ConsoleOperations {
             buf.append((char) i);
         }
 
-        // return new BufferedReader (new InputStreamReader (in)).readLine ();
+    // return new BufferedReader (new InputStreamReader (in)).readLine ();
     }
 
     /**
@@ -658,17 +771,14 @@ public class ConsoleReader implements ConsoleOperations {
         short code = keybindings[c];
 
         if (debugger != null) {
-            debug("    translated: " + (int) c + ": " + code);
+            // debug("    translated: " + (int) c + ": " + code);
         }
 
-        return new int[] { c, code };
+        return new int[]{c, code};
     }
 
     /**
      * Move up or down the history tree.
-     *
-     * @param direction
-     *            less than 0 to move up the tree, down otherwise
      */
     private final boolean moveHistory(final boolean next) throws IOException {
         if (next && !history.next()) {
@@ -706,8 +816,7 @@ public class ConsoleReader implements ConsoleOperations {
         }
 
         try {
-            Object content = transferable
-                    .getTransferData(DataFlavor.plainTextFlavor);
+            Object content = transferable.getTransferData(DataFlavor.plainTextFlavor);
 
             /*
              * This fix was suggested in bug #1060649 at
@@ -735,8 +844,7 @@ public class ConsoleReader implements ConsoleOperations {
 
                 String line = null;
 
-                for (BufferedReader read = new BufferedReader((Reader) content); (line = read
-                        .readLine()) != null;) {
+                for (BufferedReader read = new BufferedReader((Reader) content); (line = read.readLine()) != null;) {
                     if (value.length() > 0) {
                         value += "\n";
                     }
@@ -755,8 +863,9 @@ public class ConsoleReader implements ConsoleOperations {
 
             return true;
         } catch (UnsupportedFlavorException ufe) {
-            if (debugger != null)
+            if (debugger != null) {
                 debug(ufe + "");
+            }
 
             return false;
         }
@@ -794,12 +903,10 @@ public class ConsoleReader implements ConsoleOperations {
         }
 
         // send the ANSI code to clear the screen
-        printString(((char) 27) + "[2J");
-        flushConsole();
+        printANSISequence("2J");
 
         // then send the ANSI code to go to position 1,1
-        printString(((char) 27) + "[1;1H");
-        flushConsole();
+        printANSISequence("1;1H");
 
         redrawLine();
 
@@ -858,7 +965,7 @@ public class ConsoleReader implements ConsoleOperations {
         int maxwidth = 0;
 
         for (Iterator i = stuff.iterator(); i.hasNext(); maxwidth = Math.max(
-                maxwidth, i.next().toString().length())) {
+                        maxwidth, i.next().toString().length())) {
             ;
         }
 
@@ -866,10 +973,11 @@ public class ConsoleReader implements ConsoleOperations {
 
         int showLines;
 
-        if (usePagination)
+        if (usePagination) {
             showLines = getTermheight() - 1; // page limit
-        else
+        } else {
             showLines = Integer.MAX_VALUE;
+        }
 
         for (Iterator i = stuff.iterator(); i.hasNext();) {
             String cur = (String) i.next();
@@ -882,14 +990,15 @@ public class ConsoleReader implements ConsoleOperations {
                     printString(loc.getString("display-more"));
                     flushConsole();
                     int c = readVirtualKey();
-                    if (c == '\r' || c == '\n')
+                    if (c == '\r' || c == '\n') {
                         showLines = 1; // one step forward
-                    else if (c != 'q')
+                    } else if (c != 'q') {
                         showLines = getTermheight() - 1; // page forward
-
+                    }
                     back(loc.getString("display-more").length());
-                    if (c == 'q')
+                    if (c == 'q') {
                         break; // cancel
+                    }
                 }
             }
 
@@ -994,8 +1103,7 @@ public class ConsoleReader implements ConsoleOperations {
         // obtain the difference between the current buffer and the new one
         int sameIndex = 0;
 
-        for (int i = 0, l1 = buffer.length(), l2 = buf.buffer.length(); (i < l1)
-                && (i < l2); i++) {
+        for (int i = 0, l1 = buffer.length(), l2 = buf.buffer.length(); (i < l1) && (i < l2); i++) {
             if (buffer.charAt(i) == buf.buffer.charAt(i)) {
                 sameIndex++;
             } else {
@@ -1003,7 +1111,11 @@ public class ConsoleReader implements ConsoleOperations {
             }
         }
 
-        int diff = buf.buffer.length() - sameIndex;
+        int diff = buf.cursor - sameIndex;
+        if (diff < 0) { // we can't backspace here so try from the end of the buffer
+                moveToEnd();
+                diff = buf.buffer.length() - sameIndex;
+        }
 
         backspace(diff); // go back for the differences
         killLine(); // clear to the end of the line
@@ -1031,7 +1143,9 @@ public class ConsoleReader implements ConsoleOperations {
         printString(buf.buffer.toString());
 
         if (buf.length() != buf.cursor) // not at end of line
-            back(buf.length() - buf.cursor); // sync
+        {
+            back(buf.length() - buf.cursor - 1); // sync
+        }
     }
 
     /**
@@ -1096,12 +1210,10 @@ public class ConsoleReader implements ConsoleOperations {
             // no masking...
             if (mask == null) {
                 printCharacter(c);
-            }
-            // null mask: don't print anything...
+            } // null mask: don't print anything...
             else if (mask.charValue() == 0) {
                 ;
-            }
-            // otherwise print the mask...
+            } // otherwise print the mask...
             else {
                 printCharacter(mask.charValue());
             }
@@ -1119,14 +1231,24 @@ public class ConsoleReader implements ConsoleOperations {
      */
     private final void drawBuffer(final int clear) throws IOException {
         // debug ("drawBuffer: " + clear);
+        if (buf.cursor == buf.length() && clear == 0) {
+            return;
+        }
         char[] chars = buf.buffer.substring(buf.cursor).toCharArray();
-        if (mask != null)
+        if (mask != null) {
             Arrays.fill(chars, mask.charValue());
+        }
 
         printCharacters(chars);
-
         clearAhead(clear);
-        back(chars.length);
+        if (terminal.isANSISupported()) {
+            if (chars.length > 0) {
+                // don't ask, it seems to work
+                back(Math.max(chars.length - 1, 1));
+            }
+        } else {
+            back(chars.length);
+        }
         flushConsole();
     }
 
@@ -1146,6 +1268,11 @@ public class ConsoleReader implements ConsoleOperations {
             return;
         }
 
+        if (terminal.isANSISupported()) {
+            printANSISequence("J");
+            return;
+        }
+
         // debug ("clearAhead: " + num);
 
         // print blank extra characters
@@ -1166,6 +1293,22 @@ public class ConsoleReader implements ConsoleOperations {
      * Move the visual cursor backwards without modifying the buffer cursor.
      */
     private final void back(final int num) throws IOException {
+        if (num == 0) return;
+        if (terminal.isANSISupported()) {
+            int width = getTermwidth();
+            int cursor = getCursorPosition();
+            // debug("back: " + cursor + " + " + num + " on " + width);
+            int currRow = (cursor + num) / width;
+            int newRow = cursor / width;
+            int newCol = cursor % width + 1;
+            // debug("    old row: " + currRow + " new row: " + newRow);
+            if (newRow < currRow) {
+                printANSISequence((currRow - newRow) + "A");
+            }
+            printANSISequence(newCol + "G");
+            flushConsole();
+            return;
+        }
         printCharacters(BACKSPACE, num);
         flushConsole();
     }
@@ -1194,7 +1337,7 @@ public class ConsoleReader implements ConsoleOperations {
             out.write(cbuf);
             return;
         }
-	    
+
         out.write(c);
     }
 
@@ -1204,29 +1347,31 @@ public class ConsoleReader implements ConsoleOperations {
      */
     private final void printCharacters(final char[] c) throws IOException {
         int len = 0;
-        for (int i = 0; i < c.length; i++)
-            if (c[i] == '\t')
+        for (int i = 0; i < c.length; i++) {
+            if (c[i] == '\t') {
                 len += TAB_WIDTH;
-            else
+            } else {
                 len++;
+            }
+        }
 
-        char cbuf[];        
-        if (len == c.length)
+        char cbuf[];
+        if (len == c.length) {
             cbuf = c;
-        else {
+        } else {
             cbuf = new char[len];
             int pos = 0;
-            for (int i = 0; i < c.length; i++){
+            for (int i = 0; i < c.length; i++) {
                 if (c[i] == '\t') {
                     Arrays.fill(cbuf, pos, pos + TAB_WIDTH, ' ');
                     pos += TAB_WIDTH;
-	        } else {
+                } else {
                     cbuf[pos] = c[i];
                     pos++;
                 }
-	    }
+            }
         }
-            
+
         out.write(cbuf);
     }
 
@@ -1265,10 +1410,18 @@ public class ConsoleReader implements ConsoleOperations {
         }
 
         int count = 0;
-
+        int termwidth = getTermwidth();
+        int lines = getCursorPosition() / termwidth;
         count = moveCursor(-1 * num) * -1;
         // debug ("Deleting from " + buf.cursor + " for " + count);
         buf.buffer.delete(buf.cursor, buf.cursor + count);
+        if (getCursorPosition() / termwidth != lines) {
+            if (terminal.isANSISupported()) {
+                // debug("doing backspace redraw: " + getCursorPosition() + " on " + termwidth + ": " + lines);
+                printANSISequence("J");
+                flushConsole();
+            }
+        }
         drawBuffer(count);
 
         return count;
@@ -1284,15 +1437,7 @@ public class ConsoleReader implements ConsoleOperations {
     }
 
     private final boolean moveToEnd() throws IOException {
-        if (moveCursor(1) == 0) {
-            return false;
-        }
-
-        while (moveCursor(1) != 0) {
-            ;
-        }
-
-        return true;
+        return moveCursor(buf.length() - buf.cursor) > 0;
     }
 
     /**
@@ -1300,12 +1445,7 @@ public class ConsoleReader implements ConsoleOperations {
      * the buffer.
      */
     private final boolean deleteCurrentCharacter() throws IOException {
-        boolean success = buf.buffer.length() > 0;
-        if (!success) {
-            return false;
-        }
-
-        if (buf.cursor == buf.buffer.length()) {
+        if (buf.length() == 0 || buf.cursor == buf.length()) {
             return false;
         }
 
@@ -1353,20 +1493,20 @@ public class ConsoleReader implements ConsoleOperations {
     /**
      * Move the cursor <i>where</i> characters.
      *
-     * @param where
-     *            if less than 0, move abs(<i>where</i>) to the left,
-     *            otherwise move <i>where</i> to the right.
+     * @param num
+     *            if less than 0, move abs(<i>num</i>) to the left,
+     *            otherwise move <i>num</i> to the right.
      *
      * @return the number of spaces we moved
      */
     public final int moveCursor(final int num) throws IOException {
         int where = num;
 
-        if ((buf.cursor == 0) && (where < 0)) {
+        if ((buf.cursor == 0) && (where <= 0)) {
             return 0;
         }
 
-        if ((buf.cursor == buf.buffer.length()) && (where > 0)) {
+        if ((buf.cursor == buf.buffer.length()) && (where >= 0)) {
             return 0;
         }
 
@@ -1398,8 +1538,6 @@ public class ConsoleReader implements ConsoleOperations {
      * Move the cursor <i>where</i> characters, withough checking the current
      * buffer.
      *
-     * @see #where
-     *
      * @param where
      *            the number of characters to move to the right or left.
      */
@@ -1408,22 +1546,40 @@ public class ConsoleReader implements ConsoleOperations {
         // + buf.cursor + " => " + (buf.cursor + where) + ")");
         buf.cursor += where;
 
+        if (terminal.isANSISupported()) {
+            if (where < 0) {
+                back(Math.abs(where));
+            } else {
+                int width = getTermwidth();
+                int cursor = getCursorPosition();
+                int oldLine = (cursor - where) / width;
+                int newLine = cursor / width;
+                if (newLine > oldLine) {
+                    printANSISequence((newLine - oldLine) + "B");
+                }
+                printANSISequence(1 +(cursor % width) + "G");
+            }
+            flushConsole();
+            return;
+        }
+
         char c;
 
         if (where < 0) {
-	    int len = 0;
-            for (int i = buf.cursor; i < buf.cursor - where; i++){
-                if (buf.getBuffer().charAt(i) == '\t')
+            int len = 0;
+            for (int i = buf.cursor; i < buf.cursor - where; i++) {
+                if (buf.getBuffer().charAt(i) == '\t') {
                     len += TAB_WIDTH;
-                else
+                } else {
                     len++;
-	    }
+                }
+            }
 
-	    char cbuf[] = new char[len];
-	    Arrays.fill(cbuf, BACKSPACE);
-	    out.write(cbuf);
-		
-	    return;
+            char cbuf[] = new char[len];
+            Arrays.fill(cbuf, BACKSPACE);
+            out.write(cbuf);
+
+            return;
         } else if (buf.cursor == 0) {
             return;
         } else if (mask != null) {
@@ -1450,7 +1606,7 @@ public class ConsoleReader implements ConsoleOperations {
         int c = terminal.readVirtualKey(in);
 
         if (debugger != null) {
-            debug("keystroke: " + c + "");
+            // debug("keystroke: " + c + "");
         }
 
         // clear any echo characters
@@ -1466,27 +1622,24 @@ public class ConsoleReader implements ConsoleOperations {
 
         Arrays.sort(allowed); // always need to sort before binarySearch
 
-        while (Arrays.binarySearch(allowed, c = (char) readVirtualKey()) < 0)
-            ;
+        while (Arrays.binarySearch(allowed, c = (char) readVirtualKey()) < 0);
 
         return c;
     }
 
-   
     /**
      *  Issue <em>num</em> deletes.
      *
      *  @return  the number of characters backed up
      */
-    private final int delete (final int num)
-    throws IOException
-    {
-    	/* Commented out beacuse of DWA-2949:
-           if (buf.cursor == 0)
-                       return 0;*/
+    private final int delete(final int num)
+            throws IOException {
+        /* Commented out beacuse of DWA-2949:
+        if (buf.cursor == 0)
+        return 0;*/
 
-        buf.buffer.delete (buf.cursor, buf.cursor + 1);
-        drawBuffer (1);
+        buf.buffer.delete(buf.cursor, buf.cursor + 1);
+        drawBuffer(1);
 
         return 1;
     }
@@ -1509,13 +1662,11 @@ public class ConsoleReader implements ConsoleOperations {
      *
      *  @return  true if successful
      */
-    public final boolean delete ()
-    throws IOException
-    {
-        return delete (1) == 1;
+    public final boolean delete()
+            throws IOException {
+        return delete(1) == 1;
     }
 
-
     public void setHistory(final History history) {
         this.history = history;
     }
@@ -1591,6 +1742,41 @@ public class ConsoleReader implements ConsoleOperations {
         return !Character.isLetterOrDigit(c);
     }
 
+    private void printANSISequence(String sequence) throws IOException {
+        printCharacter(27);
+        printCharacter('[');
+        printString(sequence);
+        flushConsole();
+    }
+
+    /*
+    private int currentCol, currentRow;
+
+    private void getCurrentPosition() {
+        // check for ByteArrayInputStream to disable for unit tests
+        if (terminal.isANSISupported() && !(in instanceof ByteArrayInputStream)) {
+            try {
+                printANSISequence("[6n");
+                flushConsole();
+                StringBuffer b = new StringBuffer(8);
+                // position is sent as <ESC>[{ROW};{COLUMN}R
+                int r;
+                while((r = in.read()) > -1 && r != 'R') {
+                    if (r != 27 && r != '[') {
+                        b.append((char) r);
+                    }
+                }
+                String[] pos = b.toString().split(";");
+                currentRow = Integer.parseInt(pos[0]);
+                currentCol = Integer.parseInt(pos[1]);
+            } catch (Exception x) {
+                // no luck
+                currentRow = currentCol = -1;
+            }
+        }
+    }
+    */
+
     /**
      * Whether or not to add new commands to the history buffer.
      */
@@ -1621,4 +1807,17 @@ public class ConsoleReader implements ConsoleOperations {
         return this.usePagination;
     }
 
+    public void printSearchStatus(String searchTerm, String match) throws IOException {
+        int i = match.indexOf(searchTerm);
+        printString("\r(reverse-i-search) `" + searchTerm + "': " + match + "\u001b[K");
+        // FIXME: our ANSI using back() does not work here
+        printCharacters(BACKSPACE, match.length() - i);
+        flushConsole();
+    }
+
+    public void restoreLine() throws IOException {
+        printString("\u001b[2K"); // ansi/vt100 for clear whole line
+        redrawLine();
+        flushConsole();
+    }
 }
